/*
 *	usermain.c (usermain)
 *	Digital photo-frame application
 */

#include <basic.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <tk/tkernel.h>
#include <tm/tmonitor.h>
#include <device/clk.h>
#include <device/screen.h>
#include <t2ex/fs.h>
#include <t2ex/datetime.h>
#include "picojpeg.h"
#include "fileread.h"

Inline UH toRGB565(UB r, UB g, UB b)
{
	return (UH)(((r >> 3) << 11) | ((g >> 2) << 5) | (b >> 3));
}

LOCAL unsigned char pjpeg_need_bytes_callback(unsigned char* buf, unsigned char size, unsigned char *asize, void* exinf)
{
	file_reader* reader = (file_reader*)exinf;
	ER er;

	er = file_read(reader, buf, size);
	if (er < 0) {
		return PJPG_STREAM_READ_ERROR;
	}
	*asize = (unsigned char)er;
	return 0;
}

LOCAL ER   display_photo(ID scr, BMP* bmp, const char* fname)
{
	ER er;
	pjpeg_image_info_t image_info;
	unsigned char status;
	int mcu_x, mcu_y, x, y, px, py, src_ofs;
	UH* dest = (UH*)bmp->baseaddr[0];
	W asize;
	static file_reader reader;

	er = file_open(&reader, fname, O_RDONLY, 0);
	if (er < E_OK) {
		tm_putstring("file_open failed\n");
		goto err_ret0;
	}

	status = pjpeg_decode_init(&image_info, pjpeg_need_bytes_callback, &reader, FALSE);
	if (status) {
		tm_putstring("pjpeg_decode_init error\n");
		er = EX_INVAL;
		goto err_ret1;
	}

	for (mcu_x = 0, mcu_y = 0;;) {
		status = pjpeg_decode_mcu();
		if (status) {
			if (status != PJPG_NO_MORE_BLOCKS) {
				tm_putstring("pjpeg_decode_mcu error\n");
				er = EX_INVAL;
				goto err_ret1;
			}
			break;
		}

		if (mcu_y >= image_info.m_MCUSPerCol)
			break;

		for (y = 0; y < image_info.m_MCUHeight; y++) {
			py = mcu_y * image_info.m_MCUHeight + y;

			if (py > bmp->bounds.c.bottom - bmp->bounds.c.top)
				break;

			for (x = 0; x < image_info.m_MCUWidth; x++) {
				px = mcu_x * image_info.m_MCUWidth + x;
				if (px > bmp->bounds.c.right - bmp->bounds.c.left)
					break;

				/* FIXME: PJPG_GRAYSCALE未サポート */
				src_ofs = (y / 8) * 128 + (x / 8) * 64 
					+ (y % 8) * 8 + (x % 8);
				dest[(bmp->rowbytes / 2) * py + px] =
					toRGB565(image_info.m_pMCUBufR[src_ofs],
						 image_info.m_pMCUBufG[src_ofs],
						 image_info.m_pMCUBufB[src_ofs]);
			}
		}

		if (++mcu_x >= image_info.m_MCUSPerRow) {
			mcu_x = 0;
			if (++mcu_y >= image_info.m_MCUSPerCol)
				break;
		}
	}

	tk_swri_dev(scr, DN_SCRUPDRECT, &(bmp->bounds), sizeof(bmp->bounds), &asize);
	file_close(&reader);
	return E_OK;

err_ret1:
	file_close(&reader);
err_ret0:
	return er;
}

EXPORT void appl_main(void)
{
	int cnt;
	ID scr;
	W asize;
	ER er;
	DIR* dir;
	errno_t eno;
	struct dirent entry, *pentry;
	struct stat stat;
	BMP screen_bitmap;

	/* SCREENドライバのオープン */
	er = tk_opn_dev((UB*)"SCREEN", TD_UPDATE);
	if (er < E_OK) {
		tm_putstring("tk_opn_dev failed\n");
		goto err_ret0;
	}
	scr = er;

	/* SCREENドライバの属性データ DN_SCRBMP の読み出し */
	er = tk_srea_dev(scr, DN_SCRBMP, &screen_bitmap, sizeof(screen_bitmap), &asize);
	if (er < E_OK) {
		tm_putstring("tk_srea_dev failed\n");
		goto err_ret0;
	}
	if (screen_bitmap.planes != 1 || screen_bitmap.pixbits != 0x1010) {
		tm_putstring("unsupported screen bitmap format\n");
		goto err_ret0;
	}

	/* 表示内容のクリア */
	memset(screen_bitmap.baseaddr[0], 0xff, 
	       screen_bitmap.rowbytes * (screen_bitmap.bounds.c.bottom - screen_bitmap.bounds.c.top));
	tk_swri_dev(scr, DN_SCRUPDRECT, &(screen_bitmap.bounds), sizeof(screen_bitmap.bounds), &asize);

	tk_dly_tsk(1000);

	/* ファイルシステムの接続 */
	er = fs_attach("pcb0", "system", FIMP_FAT, DEV_FLAG_READONLY, NULL);
	if (er < E_OK) {
		tm_putstring("fs_attach failed\n");
		goto err_ret1;
	}

	/* 接続したファイルシステムのルートディレクトリに移動 */
	fs_chdir("/system");

	for (;;) {
		/* ディレクトリをオープン */
		dir = opendir(".");
		if (!dir) {
			tm_putstring("opendir failed\n");
			goto err_ret2;
		}

		cnt = 0;

		for (;;) {
			/* ディレクトリの中のエントリを順に取得 */
			eno = readdir_r(dir, &entry, &pentry);
			if (eno != 0) {
				tm_putstring("readdir_r failed\n");
			}
			if (!pentry) break;

#if 0
			tm_putstring("file: ");
			tm_putstring(pentry->d_name);
			tm_putstring("\n");
#endif

			/* 通常ファイル以外はスキップ */
			er = fs_stat(pentry->d_name, &stat);
			if (er < E_OK || !(stat.st_mode & S_IFREG))
				continue;

			/* JPEGファイルをオープンしてスクリーンに表示する */
			er = display_photo(scr, &screen_bitmap, pentry->d_name);
			if (er < E_OK) continue;

			cnt++;

			/* 5秒間タスクを遅延させる */
			tk_dly_tsk(5000);
		}

		if (cnt <= 0) {
			tm_putstring("no jpeg files found\n");
			goto err_ret3;			
		}

		closedir(dir);
	}

	/* ここには到達しない */
	return;

err_ret3:
	closedir(dir);
err_ret2:
	fs_detach("system");
err_ret1:
	tk_cls_dev(scr, 0);
err_ret0:
	return;
}

